#include "ConstantsCL.h"


__kernel void concatenate(	const __global uchar* compressedChunks, 
							__global uchar* outputData,
							const __global ulong* chunkLengthList,
							const __global ulong* prefixSum)
{
	// Every thread works on one chunk.

	// Calculate offset in source data.
	uint chunkByteOffset = (GIDX * BYTESPERTHREAD * 2);

	// Calculate starting position in destination buffer.
	uint outputByteOffset = (GIDX == 0? 0 : prefixSum[GIDX - 1] / 8);
	uint outputCurrentBitLength = (GIDX == 0? 0 : prefixSum[GIDX - 1] % 8);
		
	// How many bits to render?
	int remain = (int)chunkLengthList[GIDX];
	if (remain == 0)
		return;

	// Work all remaining bits.
	uint current = 0;
	while (remain > 0)
	{
		// read next input byte.
		current |= (compressedChunks[chunkByteOffset] << outputCurrentBitLength);
		chunkByteOffset++;
		
		// write next output byte.
		outputData[outputByteOffset] = (uchar)(current & 0xFF);
		outputByteOffset++;
		
		current >>= 8;

		// reduce bits to do by one byte.
		remain -= 8;
	}
	
	// Synch all threads.
	barrier(CLK_LOCAL_MEM_FENCE);

	// Fix concatenation position.
	if (outputCurrentBitLength != 0)
	//if (GIDX != 0)
	{
		chunkByteOffset = (GIDX * BYTESPERTHREAD * 2);
		current = compressedChunks[chunkByteOffset] << outputCurrentBitLength;
		
		// Load previous part of byte.
		uint previousOffset = ((GIDX - 1) * BYTESPERTHREAD * 2);
		uint previousLength = chunkLengthList[GIDX - 1];
		uint previousChunkTail = previousOffset + previousLength / 8;
		uint previousToken = 
			compressedChunks[previousChunkTail - 1] | (compressedChunks[previousChunkTail] << 8); 
		uint tailBits = 8 + previousLength % 8;
		previousToken >>= (tailBits - outputCurrentBitLength);
		previousToken &= ((1 << outputCurrentBitLength) - 1);
		current |= previousToken;
		
		outputByteOffset = prefixSum[GIDX - 1] / 8;
		outputData[outputByteOffset] = current & 0xFF;
	}
	
}


/*	// Do we need some bits from the previous chunk?
	if (outputCurrentBitLength != 0)
	{
		// Yes, take them and store them in current.
		uint previousOffset = ((GIDX - 1) * BYTESPERTHREAD * 2);
		int remainSize = (int)chunkLengthList[GIDX - 1];
		uint byteOffset = remainSize / 8;
		uint bitsOffset = remainSize % 8;
		uint tail = compressedChunks[byteOffset - 1] + (compressedChunks[byteOffset] << 8);
		int bits = 8 + bitsOffset;
		tail >>= (bits - outputCurrentBitLength);
		current = tail & ((1 << outputCurrentBitLength) - 1);
	}
	*/
