
#include <stdio.h>
#include <tchar.h>
#include <string>
#include <iostream>

#include "CPUHuffman.h"
#include "GPUHuffman.h"

#include "CTimer.h"

int _tmain(int argc, _TCHAR* argv[])
{
	try
	{
		// Initialize a CPU and GPU compression object.
		CPUHuffman codec;
		GPUHuffman codecGPU;

		// Error in stream loading/saving.
		// I don't check it further because it is not part of the task.
		//codec.HuffmanCompress::compress("GPUHuffman.h");
		//codec.HuffmanCompress::decompress("GPUHuffman.h.huf", "GPUHuffman_cmp.h");
		
		// Iterate with different sizes for compressen-decompression validation.
		for (int iterator = 1; iterator < 10; iterator++)
		{
			// Define a random seed size e.g. 10281. Feel free to change
			#define __SIZE 102810
			DataChunk data;
			data.reserve(__SIZE*iterator);
			for (__int64 i = 0; i < __SIZE * iterator; i++)
				data.push_back(32*i & 31);

			// Compress with both algorithm for comparance.
			CompressedChunk compressedCPU = codec.compress(data);
			CompressedChunk compressedGPU = codecGPU.compress(data);

			// Decompress (both decompressions are made on the CPU.
			DataChunk decompressedCPU = codec.decompress(compressedCPU);
			DataChunk decompressedGPU = codecGPU.decompress(compressedGPU);

			//std::string compareString(decompressed.begin(), decompressed.end());
			std::cout << "Decompressed Size: " << data.size() << " bytes. " << std::endl;
			std::cout << "Compressed size : " << compressedGPU.data.size() << " bytes.\n" << std::endl;
			std::cout << "Compare CPU: " << memcmp(decompressedCPU.data(), data.data(), data.size() - 1) << std::endl;
			std::cout << "Compare GPU: " << memcmp(decompressedGPU.data(), data.data(), data.size() - 1) << std::endl;
			std::cout << "-----------" << std::endl;

		}
		
		std::cout << std::endl << std::endl << "Speed test." << std::endl << "----------" << std::endl;
		
		for (int i = 0; i < 14; i++)
		{
			DataChunk data;
			size_t size = (2 << i) * 10000;
			data.reserve(size);
			for (__int64 i = 0; i < size; i++)
				data.push_back(32 * i & 31);

			std::cout << "Compression of " << size << " bytes:" << std::endl;

			double timeCPU = codec.stressTest(data);
			double timeGPU = codecGPU.stressTest(data);

			std::cout << "CPU : " << timeCPU << " seconds." << std::endl;
			std::cout << "GPU : " << timeGPU << " seconds." << std::endl;

			std::cout << "-----------" << std::endl;
		}

		codecGPU.analyzeComponents(163840000, 10);
	}
	catch (std::exception *e)
	{
		std::cout << e->what();
		exit(1);
	}
	
	//std::cin.get();
	return 0;
}

