#include "ConstantsCL.h"

#define OFFSET(__OFFS) __OFFS

__kernel void scan(const __global ulong* inArray, __global ulong* outArray, uint N, uint offset) 
{
	int GID = get_global_id(0);
	if (GID < N) {
		if(GID < offset){
			outArray[GID] = inArray[GID];
		} else {
			outArray[GID] = inArray[GID] + inArray[GID-offset];
		}
	}
}

__kernel void Scan_WorkEfficient(__global ulong* array, __global ulong* higherLevelArray, __local ulong* localBlock) 
{
	int gid = get_global_id(0);
	int lid = get_local_id(0);
	int bid = get_group_id(0);
	int gid2 = gid * 2;
	int lid2 = lid * 2;

	// Fill local memory.
	localBlock[OFFSET(lid2)]		= array[gid2];
	localBlock[OFFSET(lid2 + 1)]	= array[gid2 + 1];

	// Wait for all threads.
	barrier(CLK_LOCAL_MEM_FENCE);

	// Up-Sweep
	int offset = 1;
	int N = get_local_size(0) * 2;
	int N2 = N / 2;
	for (int i = N2; i > 0; i /= 2)
	{
		if (lid < i)
		{
			int leftOffset  = offset * (lid2 + 2) - 1;
			int rightOffset	= offset * (lid2 + 1) - 1;
			localBlock[OFFSET(leftOffset)] += localBlock[OFFSET(rightOffset)];
		}

		// Wait for all threads.
		barrier(CLK_LOCAL_MEM_FENCE);

		offset *= 2;
	}

	// Clear last element.
	if (lid == 0) 
	{
		higherLevelArray[bid] = localBlock[OFFSET(N - 1)];
		localBlock[OFFSET(N - 1)] = 0;
	}

	// Wait for all threads.
	barrier(CLK_LOCAL_MEM_FENCE);

	// Down-Sweep.
	// Use of offset variable from up-sweep.
	for (int i = 1; i < N; i *= 2)
	{
		// Wait for all threads.
		barrier(CLK_LOCAL_MEM_FENCE);

		offset /= 2;
		if (lid < i)
		{
			int leftOffset	= offset * (lid2 + 1) - 1;
			int rightOffset = offset * (lid2 + 2) - 1;
			ulong temp = localBlock[OFFSET(leftOffset)];
			localBlock[OFFSET(leftOffset)] = localBlock[OFFSET(rightOffset)];
			localBlock[OFFSET(rightOffset)] += temp;
		}
	}

	// Wait for all threads.
	barrier(CLK_LOCAL_MEM_FENCE);

	// Write back local memory.
	array[gid2]		+= localBlock[OFFSET(lid2)];
	array[gid2 + 1]	+= localBlock[OFFSET(lid2 + 1)];

}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Kernel that should add the group PPS to the local PPS (Figure 14)
__kernel void Scan_WorkEfficientAdd(__global ulong* higherLevelArray, __global ulong* array, __local ulong* localBlock) 
{
    int gid = get_global_id(0);
    int bid = get_group_id(0);
	int lid = get_local_id(0);
	int gid2 = gid * 2;

	/*
	array[gid2] = array[gid2+1] = bid;
	return;
	*/

	if (bid > 0)
	{
		if (lid == 0)
		{
			localBlock[bid] = higherLevelArray[bid - 1];
		}

		barrier(CLK_LOCAL_MEM_FENCE);

		long blockSum = localBlock[bid];

		array[gid2]		+=	blockSum;
		array[gid2 + 1]	+=	blockSum;
	}
}